/*	pintvalid.cpp

	Validation test for the pint layer of the reference C++
	library implementation of the new Parallel Bit Pattern
	computation model.

	Copyright (C) 2022 by Henry Dietz, All rights reserved.

	This library is free software available under CC BY 4.0,
	https://creativecommons.org/licenses/by/4.0/

	This version is available at http://aggregate.org/PBP
	along with any relevant documentation, such as the
	reference card and citations to publications.

	This code is distributed in the hope that it will be
	useful, but WITHOUT ANY WARRANTY; without even the
	implied warranty of MERCHANTABILITY or FITNESS FOR A
	PARTICULAR PURPOSE.
*/

// This program generates a lot of output to stderr
// if REWAYS has a large value; 8 is sufficient for
// all the demos to run correctly
#define	REWAYS	16
#define	AOBWAYS	8

#include "pbp.h"

#define	TESTBITS	7
#define	TESTVALS	(1<<TESTBITS)
#define	forTESTVALS(I)	for (int I=0; I<TESTVALS; ++I)

int	ia[TESTVALS];
int	ib[TESTVALS];
int	ic[TESTVALS];
pint	pa, pb, pc;
pint	pk;
int	testsigned = 0;
int	errs = 0;

int
check(const pint& p, const int *i)
{
	// check that p matches *i
	int t[TESTVALS];

	int r = p.Gather(&(t[0]), TESTVALS);
	if (r != TESTVALS) {
		fprintf(stderr, "Gather(,%d) returned %d\n", TESTVALS, r);
		++errs;
	}
	forTESTVALS(j) {
		if (t[j] != i[j]) {
			fprintf(stderr, "E-channel %d was %d, not %d\n", j, t[j], i[j]);
			++errs;
		}
	}
	return(errs);
}

int
checkint(const int v, const int ref)
{
	// check that v matches ref
	if (v != ref) {
		fprintf(stderr, "Value was %d, not %d\n", v, ref);
		++errs;
	}
	return(errs);
}

int
testrand(void)
{
	// make random pint values
	testsigned = 0;
	pk = pint(rand() & (TESTVALS-1));
	forTESTVALS(i) {
		ia[i] = (rand() & (TESTVALS-1));
		ib[i] = (rand() & (TESTVALS-1));
	}
	pa = Scatter(&(ia[0]), TESTVALS);
	errs += check(pa, &(ia[0]));
	pb = Scatter(&(ib[0]), TESTVALS);
	errs += check(pb, &(ib[0]));
	return(errs);
}

int
testsrand(void)
{
	// make random pint values
	testsigned = 1;
	pk = pint(-(rand() & (TESTVALS/2-1)));
	forTESTVALS(i) {
		ia[i] = (rand() & (TESTVALS-1)) - (TESTVALS/2-1);
		ib[i] = (rand() & (TESTVALS-1)) - (TESTVALS/2-1);
	}
	pa = Scatter(&(ia[0]), TESTVALS);
	errs += check(pa, &(ia[0]));
	pb = Scatter(&(ib[0]), TESTVALS);
	errs += check(pb, &(ib[0]));
	return(errs);
}

int
testLAnd(void)
{
	// test LAnd()
	pc = pa.LAnd(pb);
	forTESTVALS(i) ic[i] = ((ia[i] != 0) && (ib[i] != 0));
	return(check(pc, &(ic[0])));
}

int
testLOr(void)
{
	// test LOr()
	pc = pa.LOr(pb);
	forTESTVALS(i) ic[i] = ((ia[i] != 0) || (ib[i] != 0));
	return(check(pc, &(ic[0])));
}

int
testLXOr(void)
{
	// test LXOr()
	pc = pa.LXOr(pb);
	forTESTVALS(i) ic[i] = ((ia[i] != 0) ^ (ib[i] != 0));
	return(check(pc, &(ic[0])));
}

int
testAnd(void)
{
	// test And()
	pc = pa & pb;
	forTESTVALS(i) ic[i] = ia[i] & ib[i];
	return(check(pc, &(ic[0])));
}

int
testOr(void)
{
	// test Or()
	pc = pa | pb;
	forTESTVALS(i) ic[i] = ia[i] | ib[i];
	return(check(pc, &(ic[0])));
}

int
testXOr(void)
{
	// test XOr()
	pc = pa ^ pb;
	forTESTVALS(i) ic[i] = ia[i] ^ ib[i];
	return(check(pc, &(ic[0])));
}

int
testEQ(void)
{
	// test EQ()
	pc = (pa == pb);
	forTESTVALS(i) ic[i] = (ia[i] == ib[i]);
	return(check(pc, &(ic[0])));
}

int
testNE(void)
{
	// test NE()
	pc = (pa != pb);
	forTESTVALS(i) ic[i] = (ia[i] != ib[i]);
	return(check(pc, &(ic[0])));
}

int
testGT(void)
{
	// test GT()
	pc = (pa > pb);
	forTESTVALS(i) ic[i] = (ia[i] > ib[i]);
	return(check(pc, &(ic[0])));
}

int
testLT(void)
{
	// test LT()
	pc = (pa < pb);
	forTESTVALS(i) ic[i] = (ia[i] < ib[i]);
	return(check(pc, &(ic[0])));
}

int
testGE(void)
{
	// test GE()
	pc = (pa >= pb);
	forTESTVALS(i) ic[i] = (ia[i] >= ib[i]);
	return(check(pc, &(ic[0])));
}

int
testLE(void)
{
	// test LE()
	pc = (pa <= pb);
	forTESTVALS(i) ic[i] = (ia[i] <= ib[i]);
	return(check(pc, &(ic[0])));
}

int
testShR(void)
{
	// test ShR()
	pc = pa.ShR(pb & 3);
	forTESTVALS(i) ic[i] = (ia[i] >> (ib[i] & 3));
	return(check(pc, &(ic[0])));
}

int
testShL(void)
{
	// test ShL()
	pc = pb & pint(3);
	pc = pa.ShL(pc);
	forTESTVALS(i) ic[i] = (ia[i] << (ib[i] & 3));
	check(pc, &(ic[0]));

	// test ShL() by an int
	for (int s=0; s<3; ++s) {
		forTESTVALS(i) ic[i] = (ia[i] << s);
		check(pa.ShL(s), &(ic[0]));
	}
	return(errs);
}

int
testAdd(void)
{
	// test Add()
	pc = (pa + pb);
	forTESTVALS(i) ic[i] = (ia[i] + ib[i]);
	return(check(pc, &(ic[0])));
}

int
testSub(void)
{
	// test Sub()
	pc = (pa - pb);
	forTESTVALS(i) ic[i] = (ia[i] - ib[i]);
	return(check(pc, &(ic[0])));
}

int
testMul(void)
{
	// test Mul()
	pc = (pa * pb);
	forTESTVALS(i) ic[i] = (ia[i] * ib[i]);
	check(pc, &(ic[0]));

	if (!testsigned) {
		// test Mul() constrained to bottom bits
		int b = (rand() & 3) + 3;
		pc = pa.Mul(pb, b);
		forTESTVALS(i) ic[i] = (ia[i] * ib[i]) & ((1 << b) - 1);
		check(pc, &(ic[0]));
	}

	return(errs);
}

int
testDiv(void)
{
	// test Div()
	pc = (pa / pb);
	forTESTVALS(i) ic[i] = (ib[i] ? (ia[i] / ib[i]) : 0);
	return(check(pc, &(ic[0])));
}

int
testMod(void)
{
	// test Mod()
	pc = (pa % pb);
	forTESTVALS(i) ic[i] = (ib[i] ? (ia[i] % ib[i]) : 0);
	return(check(pc, &(ic[0])));
}

int
testLNot(void)
{
	// test LNot()
	pc = !pa;
	forTESTVALS(i) ic[i] = !ia[i];
	return(check(pc, &(ic[0])));
}

int
testNot(void)
{
	// test Not()
	pc = ~pa;
	if (testsigned) {
		forTESTVALS(i) ic[i] = ~ia[i];
	} else {
		forTESTVALS(i) ic[i] = ((TESTVALS - 1) & ~ia[i]);
	}
	return(check(pc, &(ic[0])));
}

int
testLogic(void)
{
	// test Logic()
	pc = pa.Logic();
	forTESTVALS(i) ic[i] = (ia[i] != 0);
	return(check(pc, &(ic[0])));
}

int
testNeg(void)
{
	// test Neg()
	pc = -pa;
	forTESTVALS(i) ic[i] = -ia[i];
	return(check(pc, &(ic[0])));
}

int
testAbs(void)
{
	// test Abs()
	pc = pa.Abs();
	forTESTVALS(i) ic[i] = ((ia[i] < 0) ? -ia[i] : ia[i]);
	return(check(pc, &(ic[0])));
}

int
testReset(void)
{
	// test Reset()
	int r = (rand() & (TESTVALS - 1));
	pc = pa.Reset(r);
	forTESTVALS(i) ic[i] = ia[i];
	ic[r] = 0;
	return(check(pc, &(ic[0])));
}

int
testSet(void)
{
	// test Set()
	int r = (rand() & (TESTVALS - 1));
	int v = (rand() & (TESTVALS - 1));
	pc = pa.Set(r, v);
	forTESTVALS(i) ic[i] = ia[i];
	ic[r] = v;
	return(check(pc, &(ic[0])));
}

int
testintFirst(void)
{
	// test First()

	// make ic & pc have no zeros
	forTESTVALS(i) ic[i] = ((rand() % (TESTVALS-1)) + 1);
	pc = Scatter(&(ic[0]), TESTVALS);

	checkint(pc.First(), 0);
	forTESTVALS(i) {
		pc = pc.Reset(i);
		checkint(pc.First(), ((i == (TESTVALS-1)) ? REBITS : (i+1)));
	}
	return(errs);
}

int
testintPop(void)
{
	// test Pop()
	int pop = 0;
	forTESTVALS(i) pop += (pa.Meas(i) != 0);
	checkint(pa.Pop(), pop);
	pop = 0;
	forTESTVALS(i) pop += (pb.Meas(i) != 0);
	checkint(pb.Pop(), pop);
	return(errs);
}

int
testBit(void)
{
	// test Bit()
	int b = (rand() % TESTBITS);
	pc = pa.Bit(b);
	forTESTVALS(i) ic[i] = ((ia[i] & (1 << b)) != 0);
	return(check(pc, &(ic[0])));
}

int
testMin(void)
{
	// test Min()
	pc = pa.Min(pb);
	forTESTVALS(i) ic[i] = ((ia[i] < ib[i]) ? ia[i] : ib[i]);
	return(check(pc, &(ic[0])));
}

int
testMax(void)
{
	// test Max()
	pc = pa.Max(pb);
	forTESTVALS(i) ic[i] = ((ia[i] > ib[i]) ? ia[i] : ib[i]);
	return(check(pc, &(ic[0])));
}

int
testintAny(void)
{
	// test Any()
	checkint(pint(0).Any(), 0);
	int a = 0;
	forTESTVALS(i) a |= (ia[i] != 0);
	checkint(pa.Any(), a);
	return(checkint(pint(1).Any(), 1));
}

int
testintAll(void)
{
	// test All()
	checkint(pint(0).All(), 0);
	checkint(pa.All(), 0); // tail end is all 0s
	return(checkint(pint(1).All(), 1));
}

int
testintReduceAnd(void)
{
	// test ReduceAnd() reduction
	checkint(pk.ReduceAnd(), pk.Meas());
	int v = pa.ReduceAnd();
	int r = -1;
	forTESTVALS(i) r &= ia[i];
	return(checkint(v, r));
}

int
testintReduceOr(void)
{
	// test ReduceOr() reduction
	checkint(pk.ReduceOr(), pk.Meas());
	int v = pa.ReduceOr();
	int r = 0;
	forTESTVALS(i) r |= ia[i];
	return(checkint(v, r));
}

int
testintReduceXOr(void)
{
	// test ReduceXOr() reduction
	checkint(pk.ReduceXOr(), 0);
	int v = pa.ReduceXOr();
	int r = 0;
	forTESTVALS(i) r ^= ia[i];
	return(checkint(v, r));
}

int
testintReduceAdd(void)
{
	// test ReduceAdd() reduction
	int sum = 0;
	forTESTVALS(i) sum += ia[i];
	return(checkint(pa.ReduceAdd(), sum));
}

int
testintReduceMul(void)
{
	// test ReduceMul() reduction
	pc = pint(1).Set(42, 2).Set(5, 3);
	return(checkint(pc.ReduceMul(), 6));
}

int
testintReduceMin(void)
{
	// test ReduceMin() reduction
	// ^ 15 is to make the uninitialized entanglement
	// channels not be the minimum (0)
	int min = 0 ^ 15;
	forTESTVALS(i) if ((ia[i] ^ 15) < min) min = (ia[i] ^ 15);
	pc = pa ^ pint(15);
	checkint(pc.ReduceMin(), min);
	min = 0 ^ 15;
	forTESTVALS(i) if ((ib[i] ^ 15) < min) min = (ib[i] ^ 15);
	pc = pb ^ pint(15);
	checkint(pc.ReduceMin(), min);
	return(errs);
}

int
testintReduceMax(void)
{
	// test ReduceMax() reduction
	int max = ia[0];
	forTESTVALS(i) if (ia[i] > max) max = ia[i];
	checkint(pa.ReduceMax(), max);
	max = ib[0];
	forTESTVALS(i) if (ib[i] > max) max = ib[i];
	checkint(pb.ReduceMax(), max);
	return(errs);
}

int
testScanAnd(void)
{
	// test ScanAnd()
	int p = (rand() & 7);
	pc = pint(42);
	pc = pc.Set(p, 3);
	pc = pc.ScanAnd();
	forTESTVALS(i) ic[i] = ((i < p) ? 42 : 2);
	if (p == 0) ic[0] = 3;
	check(pc, &(ic[0]));
	pc = pa.ScanAnd();
	ic[0] = ia[0];
	for (int i=1; i<TESTVALS; ++i) ic[i] = ic[i-1] & ia[i];
	return(check(pc, &(ic[0])));
}


int
testScanOr(void)
{
	// test ScanOr()
	int p = (rand() & 7);
	pc = pint(42);
	pc = pc.Set(p, 3);
	pc = pc.ScanOr();
	forTESTVALS(i) ic[i] = ((i < p) ? 42 : 43);
	if (p == 0) ic[0] = 3;
	check(pc, &(ic[0]));
	pc = pa.ScanOr();
	ic[0] = ia[0];
	for (int i=1; i<TESTVALS; ++i) ic[i] = ic[i-1] | ia[i];
	return(check(pc, &(ic[0])));
}

int
testScanXOr(void)
{
	// test ScanXOr()
	pc = pint(1).ScanXOr();
	forTESTVALS(i) ic[i] = !(i & 1);
	check(pc, &(ic[0]));
	pc = pa.ScanXOr();
	ic[0] = ia[0];
	for (int i=1; i<TESTVALS; ++i) ic[i] = ic[i-1] ^ ia[i];
	return(check(pc, &(ic[0])));
}

int
testScanAdd(void)
{
	// test ScanAdd()
	ic[0] = ia[0];
	for (int i=1; i<TESTVALS; ++i) ic[i] = (ic[i-1] + ia[i]);
	return(check(pa.ScanAdd(), &(ic[0])));
}

int
testScanMul(void)
{
	// test ScanMul()
	pc = pint(1);
	pc = pc.Set(0, ia[0]);
	int q = rand() % 8;
	pc = pc.Set(q, 0);
	forTESTVALS(i) ic[i] = 1;
	ic[0] = ia[0];
	ic[q] = 0;
	for (int i=1; i<TESTVALS; ++i) ic[i] = (ic[i-1] * ic[i]);
	return(check(pc.ScanMul(), &(ic[0])));
}

int
testScanMin(void)
{
	// test ScanMin()
	ic[0] = ia[0];
	for (int i=1; i<TESTVALS; ++i) ic[i] = ((ic[i-1] > ia[i]) ? ia[i] : ic[i-1]);
	return(check(pa.ScanMin(), &(ic[0])));
}

int
testScanMax(void)
{
	// test ScanMax()
	ic[0] = ia[0];
	for (int i=1; i<TESTVALS; ++i) ic[i] = ((ic[i-1] > ia[i]) ? ic[i-1] : ia[i]);
	return(check(pa.ScanMax(), &(ic[0])));
}

int
testintCover(void)
{
	// test Cover()
	int lo = (rand() & ((TESTVALS / 2) - 1));
	int hi = lo + (rand() & ((TESTVALS / 2) - 1));
	pc = Cover(lo, hi);
	pc.Gather(&(ic[0]), TESTVALS);
	for (int i=0; i<(hi+1-lo); ++i) {
		checkint(ic[i], i+lo);
	}
	return(errs);
}

int
testRange(void)
{
	// test Range()
	int lo = (rand() & ((TESTVALS / 2) - 1));
	int hi = lo + (rand() & ((TESTVALS / 2) - 1));
	pc = Range(lo, hi);
	forTESTVALS(i) ic[i] = ((i < (hi+1-lo)) ? (i + lo) : 0);
	return(check(pc, &(ic[0])));
}


#define	TRY(NAME) { \
	printf("Testing " #NAME ":\n"); \
	int terrs = errs; \
	if (test##NAME() > terrs) { \
		printf(#NAME " failed %d values\n", errs-terrs); \
		exit(1); \
	} else { \
		printf(#NAME " passed\n"); \
	} \
}

#define	TRYINT(NAME) { \
	printf("Testing " #NAME " reduction:\n"); \
	int terrs = errs; \
	if (testint##NAME() > terrs) { \
		printf(#NAME " failed %d values\n", errs-terrs); \
		exit(1); \
	} else { \
		printf(#NAME " passed\n"); \
	} \
}

void
tests(void)
{
	TRY(LAnd)
	TRY(LOr)
	TRY(LXOr)

	TRY(And)
	TRY(Or)
	TRY(XOr)

	TRY(EQ)
	TRY(NE)

	TRY(GT)
	TRY(LT)
	TRY(GE)
	TRY(LE)

	TRY(ShR)
	TRY(ShL)

	TRY(Add)
	TRY(Sub)

	TRY(Mul)
	TRY(Div)
	TRY(Mod)

	TRY(LNot)
	TRY(Not)
	TRY(Logic)

	TRY(Neg)
	TRY(Abs)

	TRY(Reset)
	TRY(Set)
	TRYINT(First)
	TRYINT(Pop)

	TRY(Bit)

	TRY(Min)
	TRY(Max)

	TRYINT(Any)
	TRYINT(All)
	TRYINT(ReduceAnd)
	TRYINT(ReduceOr)
	TRYINT(ReduceXOr)
	TRYINT(ReduceAdd)
	TRYINT(ReduceMul)
	TRYINT(ReduceMin)
	TRYINT(ReduceMax)

	TRY(ScanAnd)
	TRY(ScanOr)
	TRY(ScanXOr)
	TRY(ScanAdd)
//	TRY(ScanMul)
	TRY(ScanMin)
	TRY(ScanMax)

	TRYINT(Cover)
	TRY(Range)

	re.Stats(); // show lower-level statistics
}

int
main(int argc, char **argv)
{
	srand(time(0)); // for random measurements...

	TRY(rand)
	tests();
	TRY(srand)
	tests();
}
